package dbifc

import DAISY._

@daisy
class Conference {

  asAdmin {
    query("CREATE TABLE users(userID, name)")
    query("CREATE TABLE paper(paperID, confID, title)")
    query("CREATE TABLE authors(paperID, confID, authorID)")
    query("CREATE TABLE reviewer(confID, userID)")
    query("CREATE TABLE review(paperID, confID, userID, content)")
    addUser("attacker", "evilAttacker")
    addConf("confA")
    addConf("confB")
  }

  def addPaper(paperID: String, confID: String, title: String) = {
    query(s"INSERT IN paper ['$paperID', '$confID', '$title']")
  }

  def addUser(userID: String, name: String) = {
    query(s"ADD USER $userID")
    query(s"INSERT IN users ['$userID', '$name']")
    query(s"GRANT SELECT ON authors TO '$userID'")
    query(s"GRANT SELECT ON reviewer TO '$userID'")
    query(s"GRANT INSERT ON paper TO '$userID'")
  }

  def isReviewer(u: String, confID: String) = {
    val revs = select(s"{ u | EXISTS c. ((reviewer(c, u) AND c = '$confID')) }")
    Util.debugFile(s"Reviewers for $confID: $revs")
    revs.contains(Seq(u))
  }

  def addReview(paperID: String, confID: String, cont: String) = {
    val u = whoami()
    if (isReviewer(u, confID)) {
      asAdmin { query(s"INSERT IN review ['$paperID', '$confID', '$u', '$cont']") }
    } else {
      outputTo(u, "not allowed")
    }
  }

  def unauthorizedReviewInsert() = asUser("attacker") {
    addPaper("PaperA", "CCS", "TitleA")
    addReview("PaperA", "CCS", "StrongReject")
  }

  def addReviewer(confID: String, userID: String) = {
    query(s"INSERT IN reviewer ['$confID', '$userID']")
  }

  def authorizedReviewInsert() = {
    asAdmin {
      addReviewer("confA", "attacker")
    }
    asReviewer("confA")
    addReview("paperA", "confA", "StrongReject")
  }

  def addConfReviewsView(confID: String) = asAdmin {
    query(s"""CREATE VIEW confReviews$confID(paperID, userID, content) AS OWNER OF
               { p, u, c | EXISTS cnf. ((review(p, cnf, u, p) AND cnf = '$confID')) }""")
  }

  def addConfPapersView(confID: String) = asAdmin {
    query(s"""CREATE VIEW confPapers$confID(paperID, title) AS OWNER OF
                { p, t | EXISTS cnf. ((paper(p, cnf, t) AND cnf = '$confID')) }""")
  }

  def addConf(confID: String) = {
    addConfReviewsView(confID)
    addConfPapersView(confID)
  }


  def asReviewer(confID: String) = {
    val u = whoami()
    if (isReviewer(u, confID)) {
      //asAdmin { query(s"GRANT SELECT ON confReviews$confID TO $u") }
      asAdmin { query(s"GRANT SELECT ON confPapers$confID TO $u") }
    } else {
      throw new SecurityException(s"$u is not a reviewer for $confID")
    }
  }

  def noLongerReviewer(confID: String) = {
    val u = whoami()
    asAdmin { query(s"REVOKE SELECT ON confPapers$confID FROM $u") }
  }

  def leakReviewsAsAuthor() = {
    asAdmin {
      addReviewer("confA", "attacker")
      addPaper("paperA", "confA", "titleA")
    }
    val u = whoami()
    asReviewer("confA")
    val papers = select(s"{ p, t | EXISTS c. (paper(p, c, t)) }")
    noLongerReviewer("confA")
    output(papers)
  }


  def getAllPapers() = {
    asAdmin {
      addReviewer("confA", "attacker")
      addPaper("paperA", "confA", "titleA")
    }
    val u = whoami()
    asReviewer("confA")
    val papers = select(s"{ p, t | EXISTS c. (paper(p, c, t)) }")
    output(papers)
  }

  def conferencePapers(confID: String) = {
    val u = whoami()
    asReviewer(confID)
    val papers = select(s"{ p, t | confPapers$confID(p, t) }")
    outputTo(u, papers)
  }

  def getPapers() = {
    asAdmin {
      addReviewer("confA", "attacker")
      addPaper("paperA", "confA", "titleA")
    }
    conferencePapers("confA")

  }

}
