package dbifc

import prototype.queries.domain.relational.calculus._, syntax._, expressions._
import prototype.queries.operators._
import prototype.objects._, constraints._, implementation._, triggers._, events._, actions._
import prototype.statements.interfaces._, add._, changePolicy._, create._, select._,
       update._
import prototype.statements.modes._
import prototype.privileges._
import prototype.users._
import prototype.tuple.ITuple
import scala.collection.JavaConverters._


class DBUtil {

}

object DBUtil {

  def tablesInQuery(q: IQuery): Set[IType] = q match {
    case bq: BooleanQuery => tablesInExpr(bq.getExpression)
    case nbq: NonBooleanQuery => tablesInExpr(nbq.getExpression)
    case _ => Util.die(s"unhandled query type: ${q.getClass}")
  }

  def tablesInExpr(e: IExpression): Set[IType] = e match {
    case const: Constant => Set()
    case fe: FalseExpression => Set()
    case te: TrueExpression => Set()
    case ee: EqualityExpression =>
      tablesInOperand(ee.getOp1) ++ tablesInOperand(ee.getOp2)
    case pe: PolyadicExpression =>
      pe.getExpressions.asScala
        .flatMap(tablesInExpr(_))
        .toSet
    case qe: QuantifiedExpression =>
      tablesInExpr(qe.getExpression)
    case pe: PredicateExpression =>
      Set(pe.getType)
    case _ => Util.die(s"unhandled expr type: ${e.getClass}")
  }

  def tablesInOperand(op: IOperand): Set[IType] = op match {
    case v: Variable => Set()
    case c: Constant => Set()
    case _ => Util.die(s"unhandled operand type; ${op.getClass}")
  }

  def tablesInStatement(stm: IStatement): Set[IType] = stm match {
    case sel: ISelectStatement =>
      tablesInQuery(sel.getQuery)
    case upd: IUpdateStatement =>
      Set(upd.getRelationSchema)
    case _ => Util.die(s"tablesInStatement: Unhandled statement type: $stm")
  }

  def stringToUser(u: String) = DB.v.userFactory.createUser(u)

  // check whether q reads the entire contents of a table/view t:
  // Concretely, we check whether this is a query of the form
  // { (x1, .., xn) | T(x1, .., xn) } where n is the arity of T
  // and for all 1 <= i, j <= n where i != j, we have xi != xj
  // (i.e. xi and xj are distinct identifiers)
  def readsTable(q: IQuery, t: IType): Boolean = {
    q match {
      case bq: BooleanQuery => readsTable(bq.getExpression, t)
      case nbq: NonBooleanQuery => readsTable(nbq.getExpression, t)
    }
  }

  def readsTable(e: IExpression, t: IType): Boolean = {
    e match {
      case pe: PredicateExpression if pe.getType.equals(t) =>
        // check that vars are distinct
        val vars = pe.getOperands.asScala
        vars.length == vars.toSet.size
      case pe: PredicateExpression => false
      case qe: QuantifiedExpression =>
        qe.getQuantifier.isInstanceOf[ExistentialQuantifier] &&
          readsTable(qe.getExpression, t)
      case _ => false
    }
  }

  def generateSelect(t: IType): String = {
    val vars = (0 to t.getCardinality - 1) map (i => s"x_$i")
    val quant = s"${vars.mkString(", ")}"
    val pred = s"${t.getId}(${vars.mkString(", ")})"
    s"{ ${quant} | $pred }"
  }

  def unparsePrivilege(priv: IPrivilege): String = priv match {
    case ins: InsertPrivilege => "INSERT ON"
    case del: DeletePrivilege => "DELETE ON"
    case sel: SelectPrivilege => "SELECT ON"
    case crt: CreateTriggerPrivilege => "CREATE TRIGGER ON"
    case createView: CreateViewPrivilege => "CREATE VIEW"
  }

  def unparseStringSet(strings: java.util.Set[String]): String =
    strings.asScala.mkString(",")

  def unparseStringList(strings: java.util.List[String]): String =
    strings.asScala.mkString(",")

  def unparseConstraint(constraint: IConstraint): String = constraint match {
    case primKey: PrimaryKeyConstraint =>
      s"PRIMARY KEY ( ${unparseStringSet(primKey.getFields)} )"
    case unique: UniqueConstraint =>
      s"UNIQUE ( ${unparseStringSet(unique.getFields)} )"
    case foreignKey: ForeignKeyConstraint =>
      s"FOREIGN KEY ( ${unparseStringList(foreignKey.getFields)} ) REFERENCES ${foreignKey.getDestinationTable.getId} ( ${unparseStringList(foreignKey.getDestinationFields)} )"
  }

  def unparseTable(itable: ITable): String = itable match {
    case table: prototype.objects.implementation.Table =>
      val fields = table.getFields.asScala.mkString(",")
      val wit = table.getConstraints.asScala.toList match {
        case Nil => ""
        case constraints =>
          val constraintList = constraints.map(unparseConstraint).mkString(", ")
          s"WITH $constraintList"
      }
      s"${table.getId}($fields) $wit"
    case _ => Util.die(s"ITable not an instance of Table: $itable ${itable.getClass}")
  }

  def unparsePrivilegeMode(privMode: prototype.objects.PRIVILEGE_MODE) =
    privMode match {
      case PRIVILEGE_MODE.OWNER => "OWNER"
      case PRIVILEGE_MODE.INVOKER => "INVOKER"
    }
  def unparseView(iview: IView): String = iview match {
    case view: View =>
      s"CREATE VIEW ${view.getId} ( ${unparseStringList(view.getFields)} ) AS ${unparsePrivilegeMode(view.getMode)} OF ${unparseStatement(view.getSelectQuery)}"
    case _ => Util.die(s"IView instance is not a view: $iview")
  }

  def unparseTriggerEvent(triggerEvent: ITriggerEvent): String = {
    val eventType = triggerEvent.getEvent match {
      case EVENT.INSERT => "INSERT ON"
      case EVENT.DELETE => "DELETE ON"
    }
    s"$eventType ${triggerEvent.getRelationSchema.getId}"
  }

  def unparseTriggerAction(action: ITriggerAction): String = action match {
    case upd: UpdateTriggerAction =>
      val mode = upd.getMode match {
        case UPDATE_MODE.INSERT => s"INSERT IN"
        case UPDATE_MODE.DELETE => s"DELETE FROM"
      }
      val tup = unparseTuple(upd.getTuple)
      s"$mode ${upd.getRelationSchema} [ $tup ]"
    case changePol: ChangePolicyTriggerAction =>
      Util.die("policy changes in triggers not yet implemented")
  }

  def unparseTrigger(trigger: ITrigger): String = {
    s"CREATE TRIGGER ${trigger.getId} AFTER ${unparseTriggerEvent(trigger.getEvent)} " ++
      s"AS ${unparsePrivilegeMode(trigger.getMode)} " ++
      s"WHEN ${unparseExpression(trigger.getCondition)} " ++
      s"DO ${unparseTriggerAction(trigger.getAction)}"
  }

  def unparseOperand(op: IOperand): String = op match {
    case const: Constant => const.toString
    case variable: Variable => unparseVariable(variable)
  }

  def unparseOperator(op: Operator): String = op match {
    case and: AndOperator => "AND"
    case or: OrOperator => "OR"
    case all: UniversalQuantifier => "FOR ALL"
    case ex: ExistentialQuantifier => "EXISTS"
    case not: NotOperator => "NOT"
    case eq: Equality => "="
  }

  def unparseType(typ: IType): String = typ match {
    case tbl: ITable => tbl.getId
    case view: IView => view.getId
  }

  def unparseExpression(expr: IExpression): String = expr match {
    case eqExpr: EqualityExpression =>
      s"${unparseOperand(eqExpr.getOp1)} = ${unparseOperand(eqExpr.getOp2)}"
    case fals: FalseExpression => "FALSE"
    case tru: TrueExpression => "TRUE"
    case poly: PolyadicExpression =>
      poly.getExpressions.asScala
        .map(unparseExpression)
        .mkString(unparseOperator(poly.getOperator))
    case pred: PredicateExpression =>
      val arguments = pred.getOperands.asScala
        .map(unparseOperand).mkString(", ")
      s"${unparseType(pred.getType)}($arguments)"
    case quant: QuantifiedExpression =>
      val vars = unparseVarSet(quant.getVariables)
      s"${unparseOperator(quant.getQuantifier)} $vars. ( ${unparseExpression(quant.getExpression)} )"
  }

  def unparseVariable(variable: Variable): String = variable.getIdentifier

  def unparseVarList(vars: java.util.List[Variable]) = vars.asScala.map(unparseVariable).mkString(",")

  def unparseVarSet(vars: java.util.Set[Variable]) =
    // FIXME: could be more efficient
    unparseVarList(vars.asScala.toList.asJava)

  def unparseQuery(query: IQuery): String = query match {
    case bool: BooleanQuery => unparseExpression(bool.getExpression)
    case nonBool: NonBooleanQuery =>
      val varList = unparseVarList(nonBool.getOutput)
      s"{ $varList | ${unparseExpression(nonBool.getExpression)} }"
  }

  def unparseUser(user: IUser): String = user.getUsername

  def unparseTuple(tup: ITuple): String = tup.getContent.asScala
    .map(unparseOperand)
    .mkString(",")

  def unparseStatement(stm: IStatement): String = stm match {
    case addUser: IAddUser =>
      s"ADD USER ${addUser.getUser.getUsername}"
    case grant: IGrantStatement =>
      s"GRANT ${unparsePrivilege(grant.getPrivilege)} TO ${unparseUser(grant.getTarget)}"
    case revoke: IRevokeStatement =>
      s"REVOKE ${unparsePrivilege(revoke.getPrivilege)} FROM " ++
        s"${unparseUser(revoke.getTarget)}"
    case createTable: ICreateTableStatement =>
      s"CREATE TABLE ${unparseTable(createTable.getObject)}"
    case createView: ICreateViewStatement =>
      s"CREATE VIEW ${unparseView(createView.getObject)}"
    case createTrigger: ICreateTriggerStatement =>
      s"CREATE TRIGGER ${unparseTrigger(createTrigger.getObject)}"
    case select: ISelectStatement =>
      s"${unparseQuery(select.getQuery)}"
    case insert: IInsertStatement =>
      s"INSERT IN ${insert.getRelationSchema.getId} [ ${unparseTuple(insert.getTuple)} ]"
    case delete: IDeleteStatement =>
      s"DELETE FROM ${delete.getRelationSchema.getId} [ ${unparseTuple(delete.getTuple)} ]"
    case _ => Util.die(s"Unhandled query type $stm [${stm.getClass}]")
  }

}
