package dbifc

import Types._

import prototype.queries.domain.relational.calculus._, syntax._, expressions._
import prototype.queries.operators._
import prototype.objects._
import prototype.privileges._
import prototype.statements.interfaces._, select._
import sys.process._
import scala.io.Source
import scala.concurrent.SyncVar
import java.io.OutputStream

import z3.scala._

import scala.collection.JavaConverters._

/* This module defines an experimental extension to the symbolic tuples
 presented in the paper. Instead of keep track of just one conditions on
 individual tables separately, we track queries in their entirety and check
 implication between queries (which are isomorphic to FO formulas) to Z3. */
class ExternalMonitor extends MonitorConfiguration[ExternalMonitor.L] {

  var useZ3 = true

  import ExternalMonitor.L
  override def top: L = Top()

  override def bot: L = Bottom()

  def normalizeLevel(x: L): L = {
    // some heuristics:
    x match {
      case QuerySet(qs)
          if (qs.forall(q => DBUtil.tablesInQuery(q).isEmpty)) => Bottom()
      case _ => x
    }
  }

  override def join(xx: L, yy: L): L = {
    val x = normalizeLevel(xx)
    val y = normalizeLevel(yy)
    (x, y) match {
      case (Bottom(), x) => x
      case (x, Bottom()) => x
      case (Top(), _) => Top()
      case (_, Top()) => Top()
      case (QuerySet(qs1), QuerySet(qs2)) =>
        QuerySet(qs1 ++ qs2)
    }
  }

  override def meet(x: L, y: L): L = (x, y) match {
    case (Bottom(), _) => Bottom()
    case (_, Bottom()) => Bottom()
    case (Top(), x) => x
    case (x, Top()) => x
    case (QuerySet(qs1), QuerySet(qs2)) =>
      QuerySet(qs1.intersect(qs2))
  }

  val z3 = new Z3Context("MODEL" -> false)
  var predicateSymbs: Map[IType, Z3FuncDecl] = Map()

  val theSort: Z3Sort = z3.mkStringSort()

  def typeToZ3(tbl: IType): Z3FuncDecl = predicateSymbs.get(tbl) match {
    case Some(symb) => symb
    case None =>
      val fun = z3.mkFuncDecl(tbl.getId,
                              (1 to tbl.getCardinality).map(_ => theSort),
                              z3.mkBoolSort)
      predicateSymbs = predicateSymbs + (tbl -> fun)
      fun
  }

  def operandToZ3(vars: Map[String, Z3Symbol], o: IOperand): Z3AST = o match {
    case v: Variable =>
      z3.mkConst(vars.getOrElse(v.getIdentifier, Util.die(s"Unbound variable $v")), theSort)
    case c: Constant => z3.mkString(c.getValue)
  }

  def exprToZ3(vars: Map[String, Z3Symbol], e: IExpression): Z3AST = e match {
    case f: FalseExpression => z3.mkFalse
    case t: TrueExpression => z3.mkTrue
    case ee: EqualityExpression =>
      z3.mkEq(operandToZ3(vars, ee.getOp1), operandToZ3(vars, ee.getOp2))
    case pe: PolyadicExpression => pe.getOperator match {
      case andOp: AndOperator =>
        pe.getExpressions.asScala.map(exprToZ3(vars, _))
          .foldRight(z3.mkTrue)((x, y) => z3.mkAnd(x, y))
      case orOp: OrOperator =>
        pe.getExpressions.asScala.map(exprToZ3(vars, _))
          .foldRight(z3.mkFalse)((x, y) => z3.mkOr(x, y))
      case notOp: NotOperator =>
        z3.mkNot(exprToZ3(vars, pe.getExpressions.asScala.head))
    }
    case pe: PredicateExpression =>
      val fd = typeToZ3(pe.getType)
      z3.mkApp(fd, pe.getOperands.asScala.map(operandToZ3(vars, _)):_*)
    case qe: QuantifiedExpression =>
      qe.getQuantifier match {
      case ex: ExistentialQuantifier =>
        val newSymbs: Seq[(String, (Z3Symbol, Z3Sort))] =
          qe.getVariables.asScala.map(v => (v.getIdentifier, (z3.mkFreshStringSymbol(v.getIdentifier), theSort))).toList
        //val newVars: Map[String, Z3AST] =
        //  newSymbs.toMap.mapValues(x => z3.mkConst(x._1, x._2))
        val varsUpd = vars ++ newSymbs.toMap.mapValues(x => x._1)
        z3.mkExists(0, Seq(),
                    newSymbs.map(x => x._2),
                    exprToZ3(varsUpd, qe.getExpression))
    case q =>
        Util.die(s"Unhandled quantifier in Z3 conversion: $q")
    }
  }

  // Horrible hack to serialize stuff later:
  var lastQueryVars: Map[String, Z3Symbol] = Map()

  def queryVars(q: IQuery): Map[String, Z3Symbol] = q match {
    case nb: INonBooleanQuery =>
      val res = nb.getOutput.asScala.map(v => (v.getIdentifier, z3.mkFreshStringSymbol(v.getIdentifier))).toMap
      lastQueryVars = res
      res
    case b: BooleanQuery => Map()
  }

  def queryToZ3(q: IQuery): Z3AST = q match {
    case nb: INonBooleanQuery =>
      exprToZ3(queryVars(nb), nb.getExpression)
    case b: BooleanQuery =>
      lastQueryVars = Map()
      exprToZ3(Map(), b.getExpression)
  }

  def leSyntactic(qs1: Set[IQuery], qs2: Set[IQuery]): Boolean =
    qs1.map { q1 =>
      // query might be in qs2 verbatim
      qs2.contains(q1) ||
        // maybe we have all the constituent tables in qs2
        DBUtil.tablesInQuery(q1).map { t =>
          qs2.exists(q => DBUtil.readsTable(q, t))
        }.fold(false)(_||_)
    }.fold(true)(_&&_)

  /* Implication using the Z3 API: */
  /* Since the API lacks the ability to set the logic to something
   that properly handles quantifiers (which is a common case), we
   actually use `queryImpliesFile` which spawns an external Z3 process
   to do the actual check. */
  def queryImpliesAPI(q1: IQuery, q2: IQuery): Boolean = {
    val f1 = queryToZ3(q1)
    val f2 = queryToZ3(q2)
    val solver = z3.mkSolver()
    solver.push()
    solver.assertCnstr(z3.mkNot(z3.mkImplies(f1, f2)))
    val res = solver.check() match {
      case Some(b) => !b
      case None => false
    }
    Util.debugFile(s"Checking implication from $f1 to $f2: $res")
    solver.pop()
    res
  }


  /*
  def queryImpliesProc(q1: IQuery, q2: IQuery): Boolean = {
    val f1 = queryToZ3(q1)
    val v1 = lastQueryVars
    val f2 = queryToZ3(q2)
    val v2 = lastQueryVars
    val qvars = (v1.values ++ v2.values).toSet
    val solver = z3.mkSolver()
    // solver.push()
    solver.assertCnstr(z3.mkNot(z3.mkImplies(f1, f2)))

    var outputReader: java.io.BufferedReader = null
    val inputStream = new SyncVar[OutputStream]
    val io = new ProcessIO(
      stdin  => inputStream.put(stdin),
      stdout => { outputReader = new java.io.BufferedReader(new java.io.InputStreamReader(stdout)) },
      stderr => { stderr.close() })
    val proc = Seq("z3", "-in").run(io)

    def toZ3(s: String) = inputStream.get.write(s"$s\n".getBytes)

    val file = java.io.File.createTempFile("z3input", ".smt2")
    toZ3("(push)")
    val pw = new java.io.PrintWriter(file)
    Util.debugFile(predicateSymbs)
    qvars.map { x =>
      toZ3(s"(declare-fun $x () $theSort)\n")
      pw.append(s"(declare-fun $x () $theSort)\n")
    }
    predicateSymbs.values.map { decl =>
      // Util.debugFile(s"$decl")
      toZ3(s"$decl\n")
      pw.append(s"$decl\n")
    }
    solver.getAssertions.toSeq.map { asn =>
      toZ3(s"(assert $asn)\n")
      pw.append(s"(assert $asn)\n")
    }
    toZ3("(check-sat)\n")
    pw.append("(check-sat)\n")
    pw.close()

    val res = outputReader.readLine()
    val valid = res.contains("unsat")
    toZ3("(pop)")
    valid
  } */

  /* For incomprehensible reasons, the Z3 API does not allow
   setting the logic, and does not support quantifiers properly in its default
   setting; therefore, we need to actually serialize the formula to a file
   and spawn a process. This makes it way less efficient, but I couldn't find
   a way around it. */
  def queryImpliesFileSet(qs1: Set[IQuery], q2: IQuery): Boolean = {
    val solver = z3.mkSolver()
    qs1.foreach { q1 =>
      val f1 = queryToZ3(q1)
      val qvs1 = lastQueryVars
      solver.assertCnstr(z3.mkForall(0, Nil, qvs1.map(x => (x._2, theSort)).toList, f1))
    }
    val f2 = queryToZ3(q2)
    val v2 = lastQueryVars
    val qvars = (v2.values).toSet
    // solver.push()
    solver.assertCnstr(z3.mkNot(f2))
    val file = java.io.File.createTempFile("z3input", ".smt2")
    val pw = new java.io.PrintWriter(file)
    Util.debugFile(predicateSymbs)
    qvars.map { x =>
      pw.append(s"(declare-fun $x () $theSort)\n")
    }
    predicateSymbs.values.map { decl =>
      // Util.debugFile(s"$decl")
      pw.append(s"$decl\n")
    }
    solver.getAssertions.toSeq.map { asn =>
      pw.append(s"(assert $asn)\n")
    }
    pw.append("(check-sat)")
    pw.close()
    if (!qvars.isEmpty)
      Util.debugFile(file.getAbsolutePath)
    var res = ""
    val io = new ProcessIO(
      stdin  => { stdin.close() },
      stdout => { res = Source.fromInputStream(stdout).mkString("")
        stdout.close() },
      stderr => { stderr.close() })
    val proc = Seq("z3", file.getAbsolutePath).run(io)
    if (proc.exitValue() != 0) {
      Util.die(s"z3 exited with non-zero code; output: $res")
    }
    // file.close()
    val succeeded = res.contains("unsat")
    if (!succeeded) {
      Util.debugFile(s"Check failed, SMT problem in file $file")
    } else {
      file.delete()
    }
    res.contains("unsat")
  }
  /* For incomprehensible reasons, the Z3 API does not allow
   setting the logic, and does not support quantifiers properly in its default
   setting; therefore, we need to actually serialize the formula to a file
   and spawn a process. This makes it way less efficient, but I couldn't find
   a way around it. */
  def queryImpliesFile(q1: IQuery, q2: IQuery): Boolean = {
    val f1 = queryToZ3(q1)
    val v1 = lastQueryVars
    val f2 = queryToZ3(q2)
    val v2 = lastQueryVars
    val qvars = (v1.values ++ v2.values).toSet
    val solver = z3.mkSolver()
    // solver.push()
    solver.assertCnstr(z3.mkNot(z3.mkImplies(f1, f2)))
    val file = java.io.File.createTempFile("z3input", ".smt2")
    val pw = new java.io.PrintWriter(file)
    Util.debugFile(predicateSymbs)
    qvars.map { x =>
      pw.append(s"(declare-fun $x () $theSort)\n")
    }
    predicateSymbs.values.map { decl =>
      // Util.debugFile(s"$decl")
      pw.append(s"$decl\n")
    }
    solver.getAssertions.toSeq.map { asn =>
      pw.append(s"(assert $asn)\n")
    }
    pw.append("(check-sat)")
    pw.close()
    if (!qvars.isEmpty)
      Util.debugFile(file.getAbsolutePath)
    var res = ""
    val io = new ProcessIO(
      stdin  => { stdin.close() },
      stdout => { res = Source.fromInputStream(stdout).mkString("")
        stdout.close() },
      stderr => { stderr.close() })
    val proc = Seq("z3", file.getAbsolutePath).run(io)
    if (proc.exitValue() != 0) {
      Util.die(s"z3 exited with non-zero code; output: $res")
    }
    // file.close()
    val succeeded = res.contains("unsat")
    if (!succeeded) {
      Util.debugFile(s"Check failed, SMT problem in file $file")
    } else {
      file.delete()
    }
    res.contains("unsat")
  }

  def leZ3(qs1: Set[IQuery], qs2: Set[IQuery]): Boolean = {
    /*
    qs1.forall { q1 =>
      qs2.exists { q2 =>
        // queryImpliesFile(q1, q2)
        queryImpliesFile(q1, q2)
      }
     }*/
    qs1.forall { q1 =>
      qs2.exists(q2 => q1.toString() == q2.toString()) ||
      queryImpliesFileSet(qs2, q1)
    }
  }

  override def le(xx : L, yy : L): Boolean = {
    val x = normalizeLevel(xx)
    val y = normalizeLevel(yy)
    (x, y) match {
      case (a, b) if a == b => true
      case (Bottom(), _) => true
      case (_, Top()) => true
      case (Top(), Bottom()) => false
      case (QuerySet(qs), Bottom()) => false
      case (Top(), QuerySet(qs)) => false
      case (QuerySet(qs1), QuerySet(qs2)) =>
        if (useZ3) {
          leZ3(qs1, qs2)
        } else {
          leSyntactic(qs1, qs2)
        }
        // for each query, check if we can infer its value from qs2
    }
  }


  override def queryLevel(sql: SQLStmt, u: User): L = sql match {
    case is: ISelectStatement => QuerySet(Set(is.getQuery))
    case _ => Util.die(s"queryLevel called with non-select query: $sql[${sql.getClass}]")
  }

  override def userLevel(u: User): L = {
    val privs: Set[IPrivilege] = NaiveDAC.getPrivileges(u)
   val qrySet: Set[IQuery] =
      privs.flatMap { p: IPrivilege => p match {
                       case sp: SelectPrivilege =>
                         val qryString = DBUtil.generateSelect(sp.getObj)
                         val qry: IQuery = DB.v.parseQuery(u, qryString)
                           .asInstanceOf[ISelectStatement].getQuery()
                         val viewQuery = sp.getObj match {
                           case view: IView =>
                             Set(view.getSelectQuery.getQuery)
                           case _ => Set()
                         }
                         Set(qry) ++ viewQuery
                       case _ => Set().asInstanceOf[Set[IQuery]]
                     }
      }
    if (qrySet.isEmpty) {
      Bottom()
    } else {
      QuerySet(qrySet)
    }
  }

  override def cloneLevel(l: L): L = l match {
    case Top() => Top()
    case Bottom() => Bottom()
    case QuerySet(qs) => QuerySet(qs.map(q => q.copy()))
  }

  override def forgetful: Boolean = false
  override def queryVisible: Boolean = false
}

object ExternalMonitor {
  type L = DisclosureLattice
}
