package dbifc

import Types._
import Util.debug
import scala.util.matching._

import java.sql.{PreparedStatement, SQLException, Statement}
import scala.reflect.macros._
import scala.reflect.runtime.currentMirror
import scala.tools.reflect.ToolBox

import scala.collection.JavaConverters._
import prototype.database.postgresql._
import prototype.queries.domain.relational.calculus.syntax._, expressions._
import prototype.statements._, interfaces._, create._, update._, select._, changePolicy._
import prototype.objects._, constraints._
import prototype.tuple._
import prototype.results.actionResults._

class Monitor[Level](atk: User, cfg: MonitorConfiguration[Level]) {

  val universe: scala.reflect.runtime.universe.type = scala.reflect.runtime.universe
  import universe._
  val toolbox = currentMirror.mkToolBox()
  import universe.internal.reificationSupport.freshTermName

  class Labeling {
    var varMap: List[Map[Var, Level]] = List()
    // val initialPolicy = cfg.userLevel(atk)
    var pcLevel = cfg.bot
    def defaultLabel(x: Var) = cfg.bot


    def getLevel(x: Var, maps: List[Map[Var, Level]]): Level = {
      maps match {
        case Nil => defaultLabel(x)
        case labelMap :: labelMaps =>
          labelMap.get(x) match {
            case None => getLevel(x, labelMaps)
            case Some(level) => level
          }
      }
    }

    def getLevel(x: Var): Level = getLevel(x, varMap)

    def updateLevel(x: Var, l: Level, labelMaps: List[Map[Var, Level]]):
        List[Map[Var, Level]] = {
      labelMaps match {
        case Nil => Util.die("Empty label stack")
        case labelMap :: Nil => labelMap + ((x, l)) :: Nil
        case labelMap :: rest =>
          if (labelMap.contains(x)) {
            (labelMap + ((x, l))) :: rest
          } else {
            labelMap :: updateLevel(x, l, rest)
          }
      }
    }

    def setLevel(x: Var, l: Level) = {
      varMap = updateLevel(x, l, varMap)
    }


    def newLevelScope(): Unit =
      varMap = Map[Var, Level]() :: varMap

    def leaveLevelScope(): Unit =
      varMap = varMap.tail

    def resetLabels() =
      varMap = List(Map())

    def setPCLevel(l: Level) = pcLevel = l

    def getPCLevel: Level = pcLevel

    def upgradePC(l: Level) = setPCLevel(cfg.join(getPCLevel, l))

  }

  val lbls: Labeling = new Labeling()
  def pc: Level = lbls.getPCLevel

  type FunName = String
  type FunSignature = (FunName, Integer)
  type FunctionCallLevels = Map[FunSignature, List[Seq[Level]]]

  var functionCallLevels: FunctionCallLevels = Map()

  // FIXME: a lot of duplication in all these stack structures
  var functionReturnLevels: List[Level] = List()

  def bot: Level = cfg.bot

  def top: Level = cfg.top

  def newFunctionReturnScope() = {
    functionReturnLevels = cfg.bot :: functionReturnLevels
  }

  def popFunctionReturnScope(): Level = {
    // Util.debug(s"popping function return scope")
    functionReturnLevels match {
      case lev :: levels =>
        functionReturnLevels = levels
        lev
      case Nil =>
        Util.die("popFunctionReturnScope: No function return level on stack")
    }
  }

  def upgradeFunctionReturnLevel(l: Level): Unit = {
    // Util.debug(s"Upgrading return level to $l")
    functionReturnLevels match {
      case lev :: levels =>
        functionReturnLevels = cfg.join(lev, l) :: levels
      case Nil =>
        Util.warn("upgradeFunctionReturnLevel: No function return level on stack")
    }
  }

  def pushCallLevels(funSignature: FunSignature,
                     argumentLevels: Seq[Level]): Unit = {
    // Util.debug(s"Pushing call levels for $funSignature: $argumentLevels")
    functionCallLevels.get(funSignature) match {
      case Some(levelStack) =>
        functionCallLevels = functionCallLevels + ((funSignature,
                                                    (argumentLevels :: levelStack)))
      case None =>
        functionCallLevels = functionCallLevels + ((funSignature,
                                                    List(argumentLevels)))
    }
  }

  def popCallLevels(funSignature: FunSignature): Seq[Level] = {
    val res = functionCallLevels.get(funSignature) match {
      case Some(levels :: levelStack) =>
        functionCallLevels = functionCallLevels + ((funSignature, levelStack))
        levels
      case _ =>
        Util.warn(s"No function argument levels found for $funSignature")
        Seq()
    }
    // Util.debug(s"Popping call levels: $res")
    res
  }

  // The current user
  var currentUser: User = "admin"
  // var modifiedTuples: Map[ITuple, Level] = Map()

  var tableLevels: Map[IType, Level] = Map()

  def tableLevel(t: IType): Level = {
    tableLevels.get(t) match {
      case Some(l) => l
      case None =>
        val sel =
          DB.v.parseQuery(currentUser, DBUtil.generateSelect(t))
        cfg.queryLevel(sel, currentUser)
    }
  }

  def queryLevel(q: SQLStmt): Level = {
    Util.debug(s"Tables: ${DBUtil.tablesInStatement(q)}")
    Util.debug(s"Table levels: $tableLevels")
    val res = cfg.join(DBUtil.tablesInStatement(q).map(tableLevel(_)))
    Util.debug(s"query level: $q")
    res
  }

  def upgradeTableLevel(t: IType, lev: Level) = {
    Util.debug(s"Upgrading level of $t to $lev")
    val oldLevel = tableLevel(t)
    tableLevels = tableLevels + ((t, cfg.join(lev, oldLevel)))
    Util.debug(s"Table levels: $tableLevels")
  }

  def getCfg(): MonitorConfiguration[Level] = cfg

  def clearTriggers() = {
    DB.v.state.clearScheduledTriggers()
  }

  def newLevelScope(): Unit = lbls.newLevelScope()

  def leaveLevelScope(): Unit = lbls.leaveLevelScope()

  def setLevel(x: String, l: Level) = {
    lbls.setLevel(x, l)
  }

  def upgradeLevel(x: String, l: Level) = {
    lbls.setLevel(x, cfg.join(lbls.getLevel(x), l))
  }

  def upgradePC(l: Level) = {
    lbls.upgradePC(l)
  }

  def restorePC(l: Level) = {
    lbls.setPCLevel(l)
  }

  val helper = new daisyMacroHelper[universe.type]("No context information", universe)
  implicit val liftSql = Liftable[SQLStmt] { query =>
    val queryString: String = DBUtil.unparseStatement(query)
    // Util.debug(s"query sql: $queryString")
    q"dbifc.DB.v.parseQuery(${helper.monitorObject}.currentUser, $queryString)"
  }

  def expandQuery(query: SQLStmt, freeVariables: Set[String]): Tree = {
    q"${helper.monitorObject}.executeParsedQuery($query, $freeVariables)"
  }

  def runTrigger(tup: ITuple, trigger: ITrigger, freeVariables: Set[Var]) = {
    // FIXME: take invoke_mode into account
    // FIXME: probably not the right way to invoke this:
    val user = DB.v.userFactory.createUser(currentUser)
    def runAction(): QueryResult = {
      try {
        val concAction = trigger.getConcreteAction(user, tup,
                                                   DB.v.statementFactory)
        executeParsedQuery(concAction, freeVariables)
      } catch {
        case e: MalformedStatementException => e.printStackTrace()
      }
    }

    trigger.getCondition match {
      case tru: TrueExpression => runAction()
      case cond =>
        val oldPC = getPC
        newLastQueryLabel()
        val condResult =
          executeParsedQuery(trigger.getConcreteCondition(user,
                                                          tup,
                                                          DB.v.statementFactory),
                             freeVariables)
        val condLabel = popLastQueryLabel()
        upgradePC(condLabel)
        condResult match {
          case b: Boolean =>
            if (b) {
              runAction()
            }
          case set: Set[_] =>
            if (!set.isEmpty) {
              runAction()
            }
        }
        restorePC(oldPC)
    }
  }

  def runQueryStringWithExpand(queryString: String, freeVariables: Set[String]): QueryResult = {
    runQueryWithExpand(DB.v.parseQuery(currentUser, queryString), freeVariables)
  }


  def runQueryWithExpand(query: SQLStmt, freeVariables: Set[String]): QueryResult = {
    clearTriggers()
    val res = executeParsedQuery(query, freeVariables)
    query match {
      case upd: IUpdateStatement =>
        val triggers = new java.util.ArrayList(DB.v.state.getScheduledTriggers).asScala
        if (triggers.isEmpty) {
          // Util.debug(s"No triggers found for $query")
        } else {
          Util.debug(s"Found triggers: $triggers")
        }
        clearTriggers()
        triggers.foreach(runTrigger(upd.getTuple, _, freeVariables))
      case _ =>
    }
    res
  }

  var lastQueryLabel: List[Level] = List(cfg.bot)

  def upgradeLastQueryLabel(lev: Level) = {
    lastQueryLabel = lastQueryLabel match {
      case level :: levels => cfg.join(level, lev) :: levels
      case Nil =>
        Util.warn("upgradeLastQueryLabel: empty query label stack")
        Nil
    }
  }

  def newLastQueryLabel() =
    lastQueryLabel = cfg.bot :: lastQueryLabel

  def popLastQueryLabel(): Level = {
    val res = lastQueryLabel match {
      case level :: levels =>
        lastQueryLabel = levels
        level
      case Nil =>
        Util.warn("popQueryLabel: empty query label stack")
        cfg.bot
    }
    // Util.debug(s"Last query level after assignment: $res")
    res
  }

  def makeTableSelect(tbl: ITable): SQLStmt = {
    val varNames = (1 to tbl.getCardinality).map(i => s"x$i").mkString(",")
    DB.v.parseQuery(currentUser, s"{ $varNames | ${tbl.getId}($varNames) } ")
  }

  def fromITuple(tup: ITuple): Seq[String] = {
    tup.getContent.asScala.map(op => op match {
                                 case const: prototype.queries.domain.relational.calculus.syntax.expressions.Constant => const.getValue
                                 case _ => Util.die("non-constant in query result")
                               })
  }

  def executeParsedQuery(parsedQuery: SQLStmt, freeVariables: Set[String]):
      QueryResult = {
    // Executes a single query without associated triggers
    NaiveDAC.updatePolicy(currentUser, parsedQuery)
    parsedQuery match {
      case changePolicy: IChangePolicyStatement =>
        if (!cfg.le(getPC, cfg.bot)) {
          securityException(s"Policy changes not allowed in context of level ${getPC}")
        }
      case upd: IUpdateStatement =>
        // check for basic privileges to perform update:
        NaiveDAC.check(currentUser, upd) match {
          case Some(err) => securityException(err)
          case None =>
        }
        // NSU check
        val tblLevel = queryLevel(makeTableSelect(upd.getRelationSchema))
        if (!cfg.le(getPC, tblLevel)) {
          securityException(s"Inserting into table at level $tblLevel in context ${getPC}")
        } else {
          Util.debug(s"Permitting insert into table at level $tblLevel with free vars $freeVariables in context ${getPC}")
        }
        val varLevel = cfg.join(freeVariables.map(lbls.getLevel(_)))

        val tupLevel = cfg.join(getPC, varLevel)
        upgradeTableLevel(upd.getRelationSchema, tupLevel)
      case sel: ISelectStatement =>
        // sel.getQuery match {
        //   case bq: BooleanQuery =>
        //     sel.copy().setQuery(new NonBooleanQuery(bq.getExpression))
        // }
      case _ =>
        Util.warn(s"Unhandled query type: $parsedQuery [${parsedQuery.getClass}]")
    }

    Util.debug(s"executing ${parsedQuery.toSQL}")
    val result = DB.v.system.execute(parsedQuery).getActionResult
    result match {
      case singleResult: ISingleActionResult =>
        if (singleResult.exception) {
          throw new DatabaseException(singleResult.getExceptionExplanation)
        }
      case _ =>
    }
    result match {
      case nonBool: INonBooleanActionResult =>
        val res = nonBool.getResult.asScala.toSet
        // Util.debug(s"Query returned result $res")
        //upgradeLastQueryLabel(queryLevel(parsedQuery))
        upgradeLastQueryLabel(cfg.queryLevel(parsedQuery, currentUser))
        res.map(fromITuple)
      case bool: IBooleanActionResult =>
        parsedQuery match {
          case sel: ISelectStatement =>
            upgradeLastQueryLabel(cfg.queryLevel(parsedQuery, currentUser))
          case _ =>
        }
        // FIXME: Should take modifiedTuples into account
        bool.getResult
      case res =>
        Util.die(s"Unexpected result type: $res [${res.getClass}]")
    }
  }

  def query(queryString: String): QueryResult = {
    val res = executeParsedQuery(DB.v.parseQuery(currentUser, queryString), Set())
    while(!DB.v.system.noScheduledTriggers()) {
      DB.v.system.executeScheduledTrigger()
    }
    res
  }


  def getPC: Level = lbls.getPCLevel

  def getLevel(x: String): Level = lbls.getLevel(x)

  def userLevel: Level = cfg.userLevel(currentUser)

  def securityException(msg: String): Unit = {
    throw new SecurityException(s"Security exception thrown by $currentUser: $msg [current PC: ${getPC}]")
  }

  def changeUser(newUser: User) = {
    Monitor.mon.currentUser = newUser
  }

  def test() = println("Monitor test!")

  def resetLabels() = {
    lbls.resetLabels()
    restorePC(cfg.bot)
  }

}

object Monitor {
  val mon = new Monitor("atk", new ExternalMonitor())
  //val mon = new Monitor("atk", new SymbolicMonitor())
}

/* Object for interacting with the monitor. This provides
 an interface intended to be used by programs. Note however
 that invocations of some methods from here have security implications,
 such as changeUser. Just reexports methods from Monitor */
object DAISY {
  def changeUser(newUser: String) = Monitor.mon.changeUser(newUser)

  def query(queryString: String): QueryResult = {
    val res = Monitor.mon.query(queryString)
    // Util.debug(s"Result of query $queryString:\n\t$res")
    res
  }

  @daisy
  def select(queryString: String): Set[Seq[String]] = {
    query(queryString).asInstanceOf[Set[Seq[String]]]
  }

  def expandQuery(queryString: String, freeVariables: Set[String]):
      QueryResult = {
    // Util.debug(s"expandQuery got called with $queryString, $freeVariables")
    Monitor.mon.runQueryStringWithExpand(queryString, freeVariables)
  }

  def asUser[T](user: String)(code: => T): T = {
    val oldUser = Monitor.mon.currentUser
    changeUser(user)
    val res = code
    changeUser(oldUser)
    res
  }

  def asAdmin[T](code: => T): T = {
    asUser("admin") { code }
  }

  var outputs: Seq[Any] = Seq()

  // should only be called by testing code:
  def clearOutputs() = outputs = Seq()

  @sink
  def output(s: Any): Unit = {
    outputs = outputs ++ Seq(s)
    debug(s"Performing output $s")
  }

  @sink2
  def outputTo(target: User, obj: Any): Unit = {
    outputs = outputs ++ Seq((target, obj))
    debug(s"Performing output $obj to $target")
  }

  def whoami(): String = Monitor.mon.currentUser

  def run(act: => Any): Seq[Any] = {
    DB.reset()
    DAISY.changeUser("admin")
    Monitor.mon.resetLabels()
    NaiveDAC.resetPolicy()
    DAISY.clearOutputs()
    Util.withDB {
      act
    }
    DAISY.outputs
  }

}
