package dbifc

import Types._
import Util._

import java.sql.{PreparedStatement, SQLException, Statement}

import prototype.statements.interfaces._, select._, update._, create._, changePolicy._
import prototype.statements.modes._
import prototype.queries.domain.relational.calculus._, syntax._, expressions._
import prototype.objects._
import prototype.privileges._

import scala.collection.JavaConverters._

import scala.collection.mutable.{Map => MMap}

case class GrantPrivilege(p: IPrivilege) extends IPrivilege {

  def copy(): GrantPrivilege = new GrantPrivilege(p.copy)

  def toSql() = ???
}

object NaiveDAC extends DACMechanism {

  type DACResult = Option[String]

  // val selectPrivs: MMap[IType, Set[User]] = MMap()
  val privs: MMap[User, Set[IPrivilege]] = MMap()

  def usedTables(exp: IExpression): Set[IType] = exp match {
    case pe: PredicateExpression => Set(pe.getType)
    case polyExp: PolyadicExpression =>
      (polyExp.getExpressions.asScala flatMap usedTables).toSet
    case quantExp: QuantifiedExpression =>
      usedTables(quantExp.getExpression)
    case _ => Set()
  }

  def usedTables(q: IQuery): Set[IType] = q match {
    case bq: BooleanQuery => usedTables(bq.getExpression)
    case nbq: NonBooleanQuery => usedTables(nbq.getExpression)
  }

  def checkPrivileges(u: User, p: IPrivilege): DACResult = {
    if (privs.getOrElse(u, Set()).contains(p)) {
      None
    } else {
      Some(s"User $u does not have privilege $p.")
    }
  }

  def canRead(u: User, t: IType): Option[String] = {
    checkPrivileges(u, new SelectPrivilege(t))
  }

  def getPrivileges(u: User): Set[IPrivilege] =
    privs.getOrElse(u, Set())

  def addPrivilege(u: User, p: IPrivilege) =
    privs += ((u, getPrivileges(u) + p))

  def revokePrivilege(u: User, p: IPrivilege) =
    privs += ((u, getPrivileges(u) - p))

  def permitRead(u: User, t: IType): Unit =
    addPrivilege(u, new SelectPrivilege(t))

  def revokeRead(u: User, t: IType): Unit =
    revokePrivilege(u, new SelectPrivilege(t))

  def addCreationPrivileges(u: User, t: IType) = {
    val basePrivs = Seq(new SelectPrivilege(t), new InsertPrivilege(t),
                        new DeletePrivilege(t))
    basePrivs.foreach (addPrivilege(u, _))
    basePrivs.foreach (p => addPrivilege(u, GrantPrivilege(p)))
  }

  def resetPolicy() = {
    privs.clear()
  }

  override def updatePolicy(u: User, q: SQLStmt) = q match {
    case create: ICreateTableStatement =>
      addCreationPrivileges(u, create.getObject)
      None
    case createView: ICreateViewStatement =>
      addCreationPrivileges(u, createView.getObject)
      None
    case changePolicy: IChangePolicyStatement =>
      // TODO: better check for permissions:
      changePolicy.getMode match {
        case CHANGE_POLICY_MODE.GRANT =>
          addPrivilege(changePolicy.getTarget, changePolicy.getPrivilege)
        case CHANGE_POLICY_MODE.REVOKE =>
          revokePrivilege(changePolicy.getTarget, changePolicy.getPrivilege)
        case CHANGE_POLICY_MODE.GRANT_WITH_OPTION =>
          ???
      }
    case _ =>
  }

  override def check(u: User, q: SQLStmt): Option[String] = q match {
    case sel: ISelectStatement =>
      val typs = usedTables(sel.getQuery)
      typs.foldLeft(None.asInstanceOf[Option[String]])
      { (res: Option[String], t: IType) =>
        (res, canRead(u, t)) match {
          case (Some(x), None) => Some(x)
          case (None, Some(x)) => Some(x)
          case (None, None) => None
          case (Some(x), Some(y)) => Some(x ++ ", " ++ y)
        }
      }
    case ins: IInsertStatement =>
      checkPrivileges(u, new InsertPrivilege(ins.getRelationSchema))
    case changePolicy: IChangePolicyStatement =>
      // TODO: better check for permissions:
      checkPrivileges(u, GrantPrivilege(changePolicy.getPrivilege)) match {
        case Some(err) => Some(s"Policy change not permitted for $u: $err")
        case None => None
      }
    case _ =>
      // Util.warn(s"NaiveDAC: Unchecked query $q")
      None
  }
}
