package dbifc

import prototype.statements.interfaces._
import prototype.results._, actionResults._

object Types {
  type Var = String
  type Pred = Expr
  type User = String
  type Com = (User, Stmt)
  type Program = Seq[Com]
  type QueryString = String
  type QueryResult = Any //FIXME: use something more type-safe
  type SQLStmt = IStatement
  type TableName = String
}

import Types._

abstract class Value

case class IntVal(i: Integer) extends Value {
  override def toString = i.toString
}
case class StringVal(s: String) extends Value {
  override def toString = s.toString
}
case class BoolVal(b: Boolean) extends Value {
  override def toString = b.toString
}
case class SetVal(s: Set[Value]) extends Value {
  override def toString = "{" ++ s.mkString(", ") ++ "}"
}
case class TupleVal(vs: Seq[Value]) extends Value {
  override def toString = "[" ++ vs.mkString(", ") ++ "]"
}
case class DBVal(res: IActionResult) extends Value
case class UnitVal() extends Value {
  override def toString = "()"
}

trait Expr {
  def vars: Set[Var]
}

case class ValueExpr(v: Value) extends Expr {
  override def vars = Set()
}
case class VarExpr(name: Var) extends Expr {
  override def vars = Set(name)
}
case class BinExpr(op: BinOp, e1: Expr, e2: Expr) extends Expr {
  override def vars = e1.vars union e2.vars
}
case class UnExpr(op: UnOp, e: Expr) extends Expr {
  override def vars = e.vars
}

case class TupleExpr(es: Seq[Expr]) extends Expr {
  override def vars =
    (es map (_.vars)).fold(Set())((x, y) => x ++ y)
}

case class SetExpr(es: Seq[Expr]) extends Expr {
  override def vars =
    (es map (_.vars)).fold(Set())((x, y) => x ++ y)
}

abstract class BinOp

case class Plus() extends BinOp
case class Mult() extends BinOp
case class Minus() extends BinOp
case class Div() extends BinOp
case class And() extends BinOp
case class Or() extends BinOp
case class Eq() extends BinOp
case class NEq() extends BinOp
case class Le() extends BinOp
case class Ge() extends BinOp
case class Gt() extends BinOp
case class Lt() extends BinOp
case class Concat() extends BinOp
case class Intersect() extends BinOp
case class Elem() extends BinOp
case class Nth() extends BinOp

abstract class UnOp

case class Not() extends UnOp

abstract class Permission

abstract class Table

abstract class Stmt

case class Terminated() extends Stmt
case class Skip() extends Stmt
case class Assign(x: Var, e: Expr) extends Stmt
// Allows constructing a query from an expression evaluating to a string:
case class QueryStr(x: Var, e: Expr) extends Stmt
case class QueryStmt(x: Var, q: QueryString) extends Stmt
case class Out(u: User, e: Expr) extends Stmt
case class OutExpr(e1: Expr, e2: Expr) extends Stmt
case class If(e: Expr, thn: Stmt, els: Stmt) extends Stmt
case class While(e: Expr, body: Stmt) extends Stmt
case class For(x: Var, e: Expr, body: Stmt) extends Stmt
case class Throw(e: Expr) extends Stmt
case class TryCatch(body: Stmt, x: Var, handler: Stmt) extends Stmt
case class Block(stmts: Seq[Stmt]) extends Stmt

// for monitor/interpreter use only:
case class RestoreValue(x: Var, v: Value) extends Stmt
// fvs captures the free variables that were present in the original query:
case class ExpandedQuery(x: Var, q: SQLStmt, fvs: Set[Var]) extends Stmt
case class AsUser(u: User, s: Stmt) extends Stmt
case class RestorePC(l: Object) extends Stmt
case class Ex(v: Value) extends Stmt
