package dbifc

import org.scalatest._
import Types._

import prototype.tuple._
import prototype.queries.domain.relational.calculus.syntax._, expressions._

import scala.collection.JavaConverters._
import z3.scala._

class TestSpec extends FlatSpec with Matchers {
  // to turn on full stack traces in sbt session:
  // set testOptions in core += Tests.Argument("-oF")
  import DAISY.{run => runOld, _}

  @daisy
  def initCode() = {
    query("ADD USER attacker")
    query("CREATE TABLE admin(id)")
    query("""INSERT IN admin ['abc']""")
    changeUser("attacker")
    query("CREATE TABLE low(x)")
  }

  def run(act: => Any): Seq[Any] = runOld {
    initCode()
    act
  }

  def high(x: String) = Monitor.mon.setLevel(x, Monitor.mon.top)

  def shouldBeRejected(act: => Any) = {
    a [SecurityException] should be thrownBy {
      run { act }
    }
  }

  @daisy
  def lowOutput = {
    output("low")
  }
  "The basic IFC mechanism" should "accept a low outputs in low context" in {
    run { lowOutput } should be (Seq("low"))
  }

  @daisy
  def highOutput = {
    val h = 5
    high("h")
    output(h)
  }
  it should "reject high output in low context" in {
    shouldBeRejected { highOutput }
  }

  @daisy
  def lowOutputHighContext = {
    val h = 5
    high("h")
    if (h > 0) {
      output(1)
    }
  }
  it should "reject low output in high context" in {
    shouldBeRejected { lowOutputHighContext }
  }

  @daisy
  def valTest = {
    val h = 5
    high("h")
    val x = h
    output(x)
  }
  it should "reject secrets laundered through val definitions" in {
    shouldBeRejected { valTest }
  }

  @daisy
  class FuncallTest {
    def f(x: Int): Int = {
      output(x)
      x
    }

    val h = 5
    high("h")
    var y: Int = 0
    y = f(h)
    output(y)
  }
  it should "reject passing high data through id" in {
    shouldBeRejected { new FuncallTest() }
  }

  "asAdmin" should "correctly change the current user to admin" in {
    run { asAdmin { output(Monitor.mon.currentUser) } } should be (Seq("admin"))
  }

  it should "correctly switch back and forth between different users" in {
    run {
      changeUser("attacker")
      output(Monitor.mon.currentUser)
      asAdmin {
        output(Monitor.mon.currentUser)
      }
      output(Monitor.mon.currentUser)
    } should be (Seq("attacker", "admin", "attacker"))
  }


  @daisy
  def readHighTable = {
    var l: Any = null
    l = query("{x | admin(x)}")
    output(l )
  }
  "The basic query handling" should "reject outputting from high tables" in {
    shouldBeRejected { readHighTable }
  }

  @daisy
  def readHighTableVal = {
    val l = query("{x | admin(x)}")
    output(l)
  }
  "The basic query handling" should "reject laundering data from high tables through val" in {
    shouldBeRejected { readHighTableVal }
  }

  @daisy
  def userChangeCurried = {
    asUser("admin") {
      output(Monitor.mon.currentUser)
    }
  }
  it should "allow using curried functions like asUser" in {
    run { userChangeCurried } should be (Seq("admin"))
  }


  @daisy
  def readLowTable = {
    query("INSERT IN low ['someLowStuff']")
    var l: Any = null
    l = query("{ x | low(x) }")
    output(l)
  }

  @daisy
  def insertUnowned = {
    query("INSERT IN admin ['bla']")
  }
  it should "reject inserting into table without insert rights" in {
    shouldBeRejected { insertUnowned }
  }

  it should "accept outputting from a low table" in {
    run(readLowTable) should be (Seq(Set(Seq("someLowStuff"))))
  }

  @daisy
  def nsuAssignment = {
    val h = 3
    var x = 0
    high("h")
    if (h > 0) {
      x = 1
    }
  }
  it should "reject upgrading low variables in high context" in {
    shouldBeRejected { nsuAssignment }
  }

  @daisy
  def nsuInsert = {
    val h = 3
    high("h")
    if (h > 0) {
      query("INSERT IN low ['hPos']")
    }
  }
  it should "reject inserting into low table in high context" in {
    shouldBeRejected { nsuInsert }
  }


  @daisy
  def grantOnHigh = {
    changeUser("admin")
    query("GRANT SELECT ON admin TO attacker")
    changeUser("attacker")
    var x: Any = null
    x = query("{ x | admin(x) }")
    output(x)
  }
  it should "allow reading tables after a grant" in {
    run(grantOnHigh) should be (Seq(Set(Seq("abc"))))
  }

  @daisy
  def basicTrigger = {
    query("CREATE TABLE foo(y)")
    //var triggerCreation: Any = null
    query("CREATE TRIGGER insertLowTrigger AFTER INSERT ON foo AS OWNER WHEN TRUE DO INSERT IN low ['a']")
    //output(triggerCreation)
    query("INSERT IN foo ['b']")
    // var y: Any = null
    val y = select("{y | low(y)}")
    output(y)
  }

  @daisy
  def basicTrigger2 = {
    query("CREATE TABLE foo (x)")
    var triggerCreation: Any = null
    triggerCreation = query("CREATE TRIGGER insertLowTrigger AFTER INSERT ON foo AS INVOKER WHEN TRUE DO INSERT IN low [x]")
    output(triggerCreation)
    query("INSERT IN foo ['ccc']")
    var y: Any = null
    y = query("{y | low(y)}")
    output(y)
  }

  @daisy
  def basicViewAccess = {
    query("CREATE VIEW fooView (x) AS OWNER OF { y | low(y) }")
    query("INSERT IN low ['abc']")
    var res: Any = null
    res = query("{x | fooView(x)}")
    output(res)
  }
  "The view handling" should "accept basic view creation and access" in {
    run(basicViewAccess) should be (Seq(Set(Seq("abc"))))
  }

  @daisy
  def highViewGrant = {
    changeUser("admin")
    query("CREATE VIEW highView(x) AS OWNER OF { y | admin(y) }")
    query("GRANT SELECT ON highView TO attacker")
    changeUser("attacker")
    var res: Any = null
    res = query("{x | highView(x)}")
    output(res)
  }
  it should "accept reading from a high view after a grant" in {
    run(highViewGrant) should be (Seq(Set(Seq("abc"))))
  }

  @daisy
  def highViewNoGrant = {
    changeUser("admin")
    query("CREATE VIEW highView(x) AS OWNER OF { y | admin(y) }")
    changeUser("attacker")
    var res: Any = null
    res = query("{x | highView(x)}")
    output(res)
  }
  it should "reject reading a high view without permission" in {
    shouldBeRejected { highViewNoGrant }
  }

  @daisy
  def outputToHigh = {
    var res: QueryResult = null
    res = query("{x | admin(x)}")
    outputTo("admin", res)
  }
  it should "accept outputting information to a high user" in {
    run { outputToHigh } should be (Seq(("admin", Set(Seq("abc")))))
  }

  @daisy
  def outputToPublic = {
    var res: QueryResult = null
    res = query("{x | admin(x)}")
    outputTo("public", res)
  }
  it should "reject outputting high information to another low user" in {
    shouldBeRejected { outputToPublic }
  }

  "The secret table case study" should "accept the attacker reading their secrets when authorized" in {
    val result = Seq(Set(Seq("attackerSecret")))
    run { new PerUserSecrets().testAttackerAuthorized() } should be (result)
  }

  it should "reject reading your own secrets when unauthorized" in {
    shouldBeRejected { new PerUserSecrets().testAttackerUnauthorized() }
  }

  it should "reject reading others' secrets even when authorized" in {
    shouldBeRejected { new PerUserSecrets().testAttackerReadOthersAuthorized() }
  }

  it should "reject authorizing yourself as an unprivileged user" in {
    shouldBeRejected { new PerUserSecrets().testSelfAuthorize() }
  }

  "The grading case study" should "permit basic submitting and grading" in {
    val outputs = run { new Grading().submitAndGrade() }
      .head.asInstanceOf[Set[Seq[String]]].head.apply(1)
    outputs should be ("prettyGood")
  }

  it should "reject TAs accessing name-code table" in {
    shouldBeRejected { new Grading().accessStudentCodesAsTA() }
  }

  it should "accept a TA reading outputting feedback to themselves" in {
    val outputs = run { new Grading().newTAreadsFeedback() }
      .head.asInstanceOf[(String, Set[Seq[String]])]._2.head.apply(1)
    outputs should be ("prettyGood")
  }

  it should "reject TAs posting feedback publicly" in {
    shouldBeRejected { new Grading().leakFeedback() }
  }

  it should "reject students accessing other submissions" in {
    shouldBeRejected { new Grading().studentReadingOtherSubmissions() }
  }

  "The book network case study" should "permit basic reviewing and viewing your reviews" in {
    val result = Seq(Set(Seq("CritiqueOfPureReason", "PrettyDry")))
    run { new BookNetwork().basicReviewAndRead() } should be (result)
  }

  it should "accept sharing reviews to mutual friends" in {
    val result = Seq(("C", Seq("CritiqueOfPureReason", "Great")))
    run { new BookNetwork().shareReviews() } should be (result)
  }

  it should "reject leaking reviews to public" in {
    shouldBeRejected { new BookNetwork().leakReviews() }
  }

  "The calendar case study" should "accepted viewing other attendees" in {
    val result = Seq(Set("Alice", "Charlie", "Bob"))
    run { new Calendar().testAddAndView() } should be (result)
  }

  it should "reject leaking participants to public" in {
    shouldBeRejected { new Calendar().testLeakParticipants() }
  }

  it should "reject viewing participants after being removed from event" in {
    shouldBeRejected { new Calendar().viewAttendeesAfterRemoved() }
  }

  "The conference system case study" should "reject invalid review insertion" in {
    run { new Conference().unauthorizedReviewInsert() } should be
      (Seq("atk", "not allowed"))
  }

  it should "accept authorized review insertion" in {
    run { new Conference().authorizedReviewInsert() } should be (Seq())
  }

  it should "reject outputting reviews as author" in {
    shouldBeRejected { new Conference().leakReviewsAsAuthor() }
  }

  it should "allow fetching papers for reviewers of conference" in {
    val result = Seq(("attacker", Set(Seq("paperA", "titleA"))))
    run { new Conference().getPapers() } should be (result)
  }

  "The new book network example" should "allow inserting a review" in {
    val result = List(("carl",Set(Seq("a", "WarAndPeace", "Text", "Ten"))),
                      ("alice",Set(Seq("a", "WarAndPeace", "Text", "Ten"))))
    run { new BookNetworkNew().basicReviewAndRead() } should be (result)
  }

  it should "reject the same example with new trigger" in {
    val result = Seq(1)
    //run { new BookNetworkNew().reviewWithTrigger() } should be (result)
    shouldBeRejected { new BookNetworkNew().reviewWithTrigger() }
  }

  /*
  "The Z3 integration" should "show trivial implications" in {
    val z3 = new Z3Context("MODEL" -> "false")
    val s = z3.mkStringSort()
    val x = z3.mkFreshStringSymbol("x")
    val y = z3.mkFreshStringSymbol("y")
    val P =  z3.mkFuncDecl("P", Seq(s, s), z3.mkBoolSort())
    val solver = z3.mkSolver()
    solver.assertCnstr(
      z3.mkNot(
        z3.mkImplies(
          z3.mkApp(P, z3.mkConst(x, s), z3.mkConst(y, s)),
          z3.mkApp(P, z3.mkConst(x, s), z3.mkConst(y, s)))))
    solver.check() should be (Some(false))
  } */
}
